/**
 *  Java Grinder
 *  Author: Michael Kohn
 *   Email: mike@mikekohn.net
 *     Web: http://www.mikekohn.net/
 * License: GPL
 *
 * Copyright 2014-2015 by Michael Kohn
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Generator.h"
#include "Compiler.h"
#include "JavaCompiler.h"
#include "execute_static.h"
#include "ARM.h"
#include "DSPIC.h"
#include "M6502.h"
#include "C64.h"
#include "AVR8.h"
#include "Epiphany.h"
#include "MC68000.h"
#include "MIPS.h"
#include "MSP430.h"
#include "MSP430X.h"
#include "SegaGenesis.h"
#include "SNES.h"
#include "STDC.h"
#include "TI84.h"
#include "TI99.h"
#include "TMS9900.h"
#include "X86.h"
#include "X86_64.h"
#include "Z80.h"
#include "version.h"

#define STACK_LEN 65536

static Generator *new_generator(const char *chip_type)
{
  Generator *generator = NULL;

  if (strcasecmp("arm", chip_type) == 0)
  {
    generator = new ARM();
  }
    else
  if (strcasecmp("attiny13", chip_type) == 0)
  {
    generator = new AVR8(ATTINY13);
  }
    else
  if (strcasecmp("attiny85", chip_type) == 0)
  {
    generator = new AVR8(ATTINY85);
  }
    else
  if (strcasecmp("attiny84", chip_type) == 0)
  {
    generator = new AVR8(ATTINY84);
  }
    else
  if (strcasecmp("attiny2313", chip_type) == 0)
  {
    generator = new AVR8(ATTINY2313);
  }
    else
  if (strcasecmp("atmega328", chip_type) == 0)
  {
    generator = new AVR8(ATMEGA328);
  }
    else
  if (strcasecmp("atmega328p", chip_type) == 0)
  {
    generator = new AVR8(ATMEGA328P);
  }
    else
  if (strcasecmp("c64", chip_type) == 0)
  {
    generator = new C64();
  }
    else
  if (strcasecmp("dspic30f3012", chip_type) == 0)
  {
    generator = new DSPIC(DSPIC30F3012);
  }
    else
  if (strcasecmp("dspic33fj06gs101a", chip_type) == 0)
  {
    generator = new DSPIC(DSPIC33FJ06GS101A);
  }
    else
  if (strcasecmp("epiphany", chip_type) == 0)
  {
    generator = new Epiphany();
  }
    else
  if (strcasecmp("m6502", chip_type) == 0)
  {
    generator = new M6502();
  }
    else
  if (strcasecmp("mc68000", chip_type) == 0)
  {
    generator = new MC68000();
  }
    else
  if (strcasecmp("mips", chip_type) == 0)
  {
    generator = new MIPS();
  }
    else
  if (strcasecmp("msp430g2231", chip_type) == 0)
  {
    generator = new MSP430(MSP430G2231);
  }
    else
  if (strcasecmp("msp430g2452", chip_type) == 0)
  {
    generator = new MSP430(MSP430G2452);
  }
    else
  if (strcasecmp("msp430g2553", chip_type) == 0)
  {
    generator = new MSP430(MSP430G2553);
  }
    else
  if (strcasecmp("msp430f5529", chip_type) == 0)
  {
    generator = new MSP430X(MSP430F5529);
  }
    else
  if (strcasecmp("sega_genesis", chip_type) == 0)
  {
    generator = new SegaGenesis();
  }
    else
  if (strcasecmp("snes", chip_type) == 0)
  {
    generator = new SNES();
  }
    else
  if (strcasecmp("stdc", chip_type) == 0)
  {
    generator = new STDC();
  }
    else
  if (strcasecmp("ti84plus", chip_type) == 0)
  {
    generator = new TI84(TI84_PLUS);
  }
    else
  if (strcasecmp("ti84plusc", chip_type) == 0)
  {
    generator = new TI84(TI84_PLUS_C);
  }
    else
  if (strcasecmp("ti99", chip_type) == 0)
  {
    generator = new TI99();
  }
    else
  if (strcasecmp("tms9900", chip_type) == 0)
  {
    generator = new TMS9900();
  }
    else
  if (strcasecmp("x86", chip_type) == 0)
  {
    generator = new X86();
  }
    else
  if (strcasecmp("x86_64", chip_type) == 0)
  {
    generator = new X86();
  }
    else
  if (strcasecmp("z80", chip_type) == 0)
  {
    generator = new Z80();
  }

  return generator;
}

int main(int argc, char *argv[])
{
  Generator *generator;
  Compiler *compiler;
  const char *java_file;
  const char *asm_file;
  const char *chip_type;
  int option = 0;
  int n;

  printf("\nJava Grinder\n"
         "Authors: Michael Kohn, Joe Davisson\n"
         "    Web: http://www.mikekohn.net/micro/java_grinder.php\n"
         "  Email: mike@mikekohn.net\n\n"
         "Version: "VERSION"\n\n");

  if (argc < 4)
  {
    printf("Usage: %s <class> <outfile> <platform>\n"
           "   platforms:\n"
           "     attiny2313, atmega328, atmega328p, attiny85, attiny84, attiny13,\n"
           "     dspic,\n"
           "     msp430g2231, msp430g2553\n"
           "     m6502, c64\n"
           "     ti84plus\n"
           "     ti99\n", argv[0]);
    exit(0);
  }

  // Can we do .NET too? :)
  compiler = new JavaCompiler();

  for (n = 1; n < argc; n++)
  {
    if (strcmp(argv[n], "-O0") == 0)
    {
      compiler->disable_optimizer();
      continue;
    }
      else
    if (option == 0)
    {
      java_file = argv[n];
    }
      else
    if (option == 1)
    {
      asm_file = argv[n];
    }
      else
    if (option == 2)
    {
      chip_type = argv[n];
    }

    option++;
  }

  generator = new_generator(chip_type);

  if (generator == NULL)
  {
    printf("Unknown cpu type: %s\n", chip_type);
    exit(1);
  }

  if (generator->open(asm_file) == -1)
  {
    delete generator;
    exit(1);
  }

  compiler->set_generator(generator);

  if (compiler->load_class(java_file) == -1)
  {
    printf("Couldn't open class file '%s'\n", java_file);
    delete generator;
    exit(1);
  }

  int ret = 0;

  compiler->insert_static_field_defines();
  compiler->init_heap();

  do
  {
    if (compiler->add_static_initializers() == -1) { ret = -1; break; }
    // Add the main function directly under init to save a jmp.
    if (compiler->compile_methods(true) == -1) { ret = -1; break; }
    // Compile all other methods.
    if (compiler->compile_methods(false) == -1) { ret = -1; break; }
    // Add constants at end if needed.
    if (compiler->add_constants() == -1) { ret = -1; break; }
  } while(0);

  delete generator;
  delete compiler;

  return ret;
}


